import contextlib
import datetime
import logging
import threading
import time
import typing
import uuid
from typing import Optional

from fastapi.utils import is_body_allowed_for_status_code
from starlette.responses import Response

from genai._utils.general import cast_list
from genai._utils.responses import BaseErrorResponse, get_api_error_class_by_status_code
from genai.credentials import Credentials
from genai.extensions.localserver.local_base_model import LocalModel
from genai.extensions.localserver.schema import (
    TextGenerationCreateRequest,
    TextTokenizationCreateRequest,
)
from genai.schema import (
    ConcurrencyLimit,
    InternalServerErrorResponse,
    NotFoundResponse,
    TextGenerationCreateResponse,
    TextGenerationLimit,
    TextGenerationLimitRetrieveResponse,
    TextTokenizationCreateResponse,
    UnauthorizedResponse,
)

try:
    import uvicorn
    from fastapi import APIRouter, FastAPI, HTTPException, Request, status
    from fastapi.responses import JSONResponse
    from starlette.middleware.base import (
        BaseHTTPMiddleware,
        DispatchFunction,
        RequestResponseEndpoint,
    )
except ImportError as iex:
    raise ImportError(f"Could not import {iex.name}: Please install ibm-generative-ai[localserver] extension.")  # noqa: B904

logger = logging.getLogger(__name__)


class ApiErrorMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: FastAPI, dispatch: typing.Optional[DispatchFunction] = None) -> None:
        super().__init__(app, dispatch)
        app.add_exception_handler(HTTPException, self._handle_exception)

    async def _handle_exception(self, request: Request, exc: Exception):
        detail = str(exc)
        status_code = getattr(exc, "status_code", 500)
        headers = getattr(exc, "headers", None)

        if isinstance(exc, HTTPException):
            detail = exc.detail
            status_code = exc.status_code

        if not is_body_allowed_for_status_code(status_code):
            return Response(status_code=status_code, headers=headers)

        response_body: BaseErrorResponse
        try:
            cls: type[BaseErrorResponse] = get_api_error_class_by_status_code(status_code) or BaseErrorResponse
            response_body = cls.model_validate(
                cls(
                    error=detail,
                    message=detail,
                    status_code=status_code,
                    extensions={},  # type: ignore
                )
                if isinstance(detail, str)
                else cls.model_validate(detail)
            )
        except:  # noqa
            response_body = InternalServerErrorResponse(message=detail, error=detail, extensions={})  # type: ignore

        return JSONResponse(
            status_code=status_code,
            headers=headers,
            content=response_body.model_dump(),
        )

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        try:
            return await call_next(request)
        except Exception as exc:
            return await self._handle_exception(request, exc)


class ApiAuthMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: FastAPI, api_key: Optional[str] = None, insecure: bool = False):
        super().__init__(app)
        self.api_key = api_key
        self.insecure = insecure

    async def dispatch(self, request: Request, call_next):
        auth_header = request.headers["Authorization"]
        logging.debug(f"Incoming Request: Auth: {auth_header}, Endpoint: {request.url}")
        if self.insecure is False:
            if str(self.api_key) not in auth_header:
                logger.warning(f"A client used an incorrect API Key: {auth_header}")
                response_obj = UnauthorizedResponse(
                    error="Unauthorized",
                    message="API key not found",
                    extensions={},  # type: ignore
                )
                logger.debug(response_obj.model_dump())
                return JSONResponse(
                    content=response_obj.model_dump(),
                    status_code=status.HTTP_401_UNAUTHORIZED,
                )
        logging.debug("Calling next in chain")
        response = await call_next(request)
        logging.debug("Returning response")
        return response


class LocalLLMServer:
    def __init__(
        self,
        models: list[type[LocalModel]],
        port: int = 8080,
        interface: str = "0.0.0.0",
        api_key: Optional[str] = None,
        insecure_api: bool = False,
    ):
        self.models = {model.model_id: model() for model in models}
        self.port = port
        self.interface = interface
        self.insecure_api = insecure_api

        # Set the API Key
        self.api_key: str = api_key or "test" if insecure_api else str(uuid.uuid4())

        self.app = FastAPI(
            title="IBM Generative AI Local Model Server",
            debug=False,
            openapi_url=None,
            docs_url=None,
            redoc_url=None,
        )

        self.app.add_middleware(ApiErrorMiddleware)
        self.router = APIRouter()
        self.router.add_api_route("/v2/text/tokenization", self._route_text_tokenization, methods=["POST"])
        self.router.add_api_route("/v2/text/generation", self._route_text_generation, methods=["POST"])
        self.router.add_api_route("/v2/text/generation/limits", self._route_generation_limits, methods=["GET"])
        self.app.include_router(self.router)
        self.app.add_middleware(ApiAuthMiddleware, api_key=self.api_key, insecure=insecure_api)

        self.uvicorn_config = uvicorn.Config(self.app, host=interface, port=port, log_level="error", access_log=False)
        self.uvicorn_server = uvicorn.Server(self.uvicorn_config)
        self.endpoint = f"http://{self.interface}:{self.port}"
        logger.debug(f"{__name__}: API: {self.endpoint}, Insecure: {insecure_api}")

    @contextlib.contextmanager
    def run_locally(self):
        thread = threading.Thread(target=self.start_server)
        thread.start()
        try:
            while not self.uvicorn_server.started:
                time.sleep(1e-3)
            yield
        finally:
            self.uvicorn_server.should_exit = True
            thread.join()

    def get_credentials(self):
        return Credentials(api_key=self.api_key, api_endpoint=self.endpoint)

    def start_server(self):
        logger.debug("Starting server background process")

        logger.info(f"Credentials: Endpoint: {self.endpoint} API Key: {self.api_key}")
        logger.info(
            f"In Python Script: credentials = Credentials(api_key={self.api_key}, api_endpoint={self.endpoint})"
        )
        self.uvicorn_server.run()

    def stop_server(self):
        logger.debug("Stopping server background process")
        self.uvicorn_server.should_exit = True
        while not self.uvicorn_server.started:
            time.sleep(1e-3)

    def _get_model(self, model_id) -> LocalModel:
        model = self.models.get(model_id)
        if not model:
            raise HTTPException(
                status_code=404,
                detail=NotFoundResponse(
                    message="Model not found",
                    error="Not Found",
                    extensions={"state": {"model_id": model_id}},  # type: ignore
                ).model_dump(),
            )
        return model

    async def _route_text_tokenization(self, body: TextTokenizationCreateRequest) -> TextTokenizationCreateResponse:
        logger.info(f"Tokenize called: {body}")

        model = self._get_model(body.model_id)
        results = [model.tokenize(input_text=input, parameters=body.parameters) for input in cast_list(body.input)]

        return TextTokenizationCreateResponse(
            model_id=body.model_id,
            created_at=datetime.datetime.now().isoformat(),
            results=results,
        )

    async def _route_text_generation(self, body: TextGenerationCreateRequest) -> TextGenerationCreateResponse:
        logger.info(f"Text Generation Called: Model: {body.model_id}, Input: {body.input}")

        model = self._get_model(body.model_id)
        results = [model.generate(input_text=input, parameters=body.parameters) for input in cast_list(body.input)]
        return TextGenerationCreateResponse(
            id=str(uuid.uuid4()),
            model_id=body.model_id,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            results=results,
        )

    async def _route_generation_limits(self):
        logger.info("Generate Limits Called")
        response = TextGenerationLimitRetrieveResponse(
            result=TextGenerationLimit(concurrency=ConcurrencyLimit(limit=100, remaining=100))
        )
        return response



import logging

try:
    from transformers import Agent
except ImportError:
    raise ImportError(  # noqa: B904
        "Could not import HuggingFace transformers: Please install ibm-generative-ai[huggingface] extension."
    )


from typing import Optional

from genai._utils.general import to_model_instance
from genai.client import Client
from genai.schema import TextGenerationParameters

logger = logging.getLogger(__name__)


class IBMGenAIAgent(Agent):
    def __init__(
        self,
        client: Client,
        model: Optional[str] = None,
        parameters: Optional[TextGenerationParameters] = None,
        chat_prompt_template: Optional[str] = None,
        run_prompt_template: Optional[str] = None,
        additional_tools: Optional[list[str]] = None,
    ):
        super().__init__(
            chat_prompt_template=chat_prompt_template,
            run_prompt_template=run_prompt_template,
            additional_tools=additional_tools,
        )
        self.client = client
        self.model = model
        self.parameters = parameters

    def generate_one(self, prompt: str, stop: Optional[list[str]] = None):
        return self._generate([prompt], stop)[0]

    def generate_many(self, prompts: list[str], stop: Optional[list[str]] = None):
        return self._generate(prompts, stop)

    def _generate(self, prompts: list[str], stop: Optional[list[str]] = None) -> list[str]:
        final_results: list[str] = []
        if len(prompts) == 0:
            return final_results

        params = to_model_instance(self.parameters, TextGenerationParameters)
        params.stop_sequences = stop or params.stop_sequences
        for response in self.client.text.generation.create(model_id=self.model, inputs=prompts, parameters=params):
            for result in response.results:
                generated_text = result.generated_text or ""

                logger.info("Output of GENAI call: {}".format(generated_text))
                final_results.append(generated_text)

        return final_results



import os
import re
from typing import Optional

from pydantic import BaseModel, Field, SecretStr, field_validator


class Credentials(BaseModel):
    """
    The `Credentials` class represents credentials required for accessing the GENAI API.

    Attributes:
        - `api_key`: API key which can be retrieved from the UI.
        - `api_endpoint` (optional): A string representing the GENAI API endpoint (default is BAM).

    Examples:
        Create a Credentials instance with explicit api_endpoint::

            credentials = Credentials(api_key="your_api_key", api_endpoint="https://bam-api.res.ibm.com")

        Create a Credentials instance with default api_endpoint::

            credentials = Credentials(api_key="your_api_key")

        Create a Credentials instance from environment variables::

            credentials = Credentials.from_env()
    """

    api_key: SecretStr = Field(..., description="The GENAI API Key")
    api_endpoint: str = Field(..., description="GENAI API Endpoint", min_length=1)

    def __init__(self, api_key: str, api_endpoint: Optional[str] = None, **kwargs):
        api_endpoint = api_endpoint or "https://bam-api.res.ibm.com"

        if api_key is None:
            raise ValueError("api_key must be provided")

        super().__init__(api_key=api_key, api_endpoint=api_endpoint, **kwargs)

    @field_validator("api_endpoint", mode="after")
    def _validate_api_endpoint(cls, value: str):
        [api, *version] = re.split(r"(/v\d+$)", value.rstrip("/"), maxsplit=1)
        if version:
            raise ValueError(
                f"The 'api_endpoint' property should not contain any explicit API version"
                f"(rename it from '{value}' to just '{api}')"
            )
        return api

    @classmethod
    def from_env(cls, *, api_key_name: Optional[str] = None, api_endpoint_name: Optional[str] = None) -> "Credentials":
        return cls(
            api_key=os.environ[api_key_name or "GENAI_KEY"],
            api_endpoint=os.environ.get(api_endpoint_name or "GENAI_API", None),
        )



from datetime import datetime
from typing import Optional, TypeVar, Union

from pydantic import BaseModel

from genai._types import EnumLike
from genai._utils.general import cast_list, to_enum, to_enum_optional
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai._utils.validators import assert_is_not_empty_string
from genai.schema import (
    RequestApiVersion,
    RequestChatConversationIdRetrieveResponse,
    RequestEndpoint,
    RequestOrigin,
    RequestRetrieveResponse,
    RequestStatus,
)
from genai.schema._api import (
    _RequestChatConversationIdDeleteParametersQuery,
    _RequestChatConversationIdRetrieveParametersQuery,
    _RequestIdDeleteParametersQuery,
    _RequestRetrieveParametersQuery,
)
from genai.schema._endpoints import (
    RequestChatConversationIdDeleteEndpoint,
    RequestChatConversationIdRetrieveEndpoint,
    RequestIdDeleteEndpoint,
    RequestRetrieveEndpoint,
)

T = TypeVar("T", bound=BaseModel)

__all__ = ["RequestService"]


class RequestService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=RequestChatConversationIdRetrieveEndpoint)
    def chat(self, conversation_id: str) -> RequestChatConversationIdRetrieveResponse:
        """
        Args:
            conversation_id: The ID of the conversation to retrieve.

        Raises:
            ValueError: If the ID is an empty string.
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        assert_is_not_empty_string(conversation_id)
        self._log_method_execution("Requests Chat", conversation_id=conversation_id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.chat)
            http_response = client.get(
                url=self._get_endpoint(metadata.endpoint, conversationId=conversation_id),
                params=_RequestChatConversationIdRetrieveParametersQuery().model_dump(),
            )
            return RequestChatConversationIdRetrieveResponse(**http_response.json())

    @set_service_action_metadata(endpoint=RequestChatConversationIdDeleteEndpoint)
    def chat_delete(self, conversation_id: str) -> None:
        """
        Args:
            conversation_id: The ID of the conversation to delete.

        Raises:
            ValueError: If the ID is an empty string.
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        assert_is_not_empty_string(conversation_id)
        self._log_method_execution("Request Delete Chat", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.chat_delete)
            client.delete(
                url=self._get_endpoint(metadata.endpoint, conversationId=conversation_id),
                params=_RequestChatConversationIdDeleteParametersQuery().model_dump(),
            )

    @set_service_action_metadata(endpoint=RequestRetrieveEndpoint)
    def list(
        self,
        *,
        api: Optional[RequestApiVersion] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        status: Optional[EnumLike[RequestStatus]] = None,
        origin: Optional[EnumLike[RequestOrigin]] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        endpoint: Optional[Union[RequestEndpoint, list[RequestEndpoint]]] = None,
        date: Optional[datetime] = None,
    ) -> RequestRetrieveResponse:
        """
        Lists requests based on the given parameters.

        Raises:
            ValueError: If the ID is an empty string.
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        params = _RequestRetrieveParametersQuery(
            limit=limit,
            offset=offset,
            status=to_enum_optional(status, RequestStatus),
            origin=to_enum_optional(origin, RequestOrigin),
            before=before,
            after=after,
            api=to_enum_optional(api, RequestApiVersion),
            date=date,
            endpoint=(
                [to_enum(RequestEndpoint, e) for e in cast_list(endpoint) if e is not None] if endpoint else None
            ),
        ).model_dump()

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.list)
            http_response = client.get(url=self._get_endpoint(metadata.endpoint), params=params)
            return RequestRetrieveResponse(**http_response.json())

    @set_service_action_metadata(endpoint=RequestIdDeleteEndpoint)
    def delete(self, id: str) -> None:
        """
        Deletes request with the given ID.

        Raises:
            ValueError: If the ID is an empty string.
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        assert_is_not_empty_string(id)
        self._log_method_execution("Request Delete", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.delete)
            client.delete(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_RequestIdDeleteParametersQuery().model_dump(),
            )



from typing import Optional

from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai._utils.validators import assert_is_not_empty_string
from genai.schema import ModelIdRetrieveResponse, ModelRetrieveResponse
from genai.schema._api import _ModelIdRetrieveParametersQuery, _ModelRetrieveParametersQuery
from genai.schema._endpoints import ModelIdRetrieveEndpoint, ModelRetrieveEndpoint

__all__ = ["ModelService"]


class ModelService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=ModelIdRetrieveEndpoint)
    def retrieve(
        self,
        id: str,
    ) -> ModelIdRetrieveResponse:
        """
        Raises:
            ValueError: If the id parameter is an empty string.
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        assert_is_not_empty_string(id)
        self._log_method_execution("Models Retrieve", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.retrieve)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_ModelIdRetrieveParametersQuery().model_dump(),
            )
            return ModelIdRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=ModelRetrieveEndpoint)
    def list(self, *, limit: Optional[int] = None, offset: Optional[int] = None) -> ModelRetrieveResponse:
        """
        Args:
            limit: The maximum number of models to retrieve.
            offset: The number of models to skip before starting to retrieve.

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        request_parameters = _ModelRetrieveParametersQuery(limit=limit, offset=offset).model_dump()
        self._log_method_execution("Models List", **request_parameters)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.list)
            response = client.get(url=self._get_endpoint(metadata.endpoint), params=request_parameters)
            return ModelRetrieveResponse(**response.json())



from queue import Queue


class FlushableQueue(Queue):
    def flush(self):
        """
        Flushes the contents of the queue by removing all items.

        This method continuously removes items from the queue until it is empty. Each item is retrieved from the
        queue using the `get()` method, and then the `task_done()` method is called to indicate
        that the item has been processed.
        """
        while not self.empty():
            _ = self.get()
            self.task_done()



import asyncio
import threading
from asyncio import CancelledError
from contextlib import suppress
from typing import Callable, Optional

from genai._utils.shared_instance import SharedResource

__all__ = ["shared_event_loop"]


class _SharedEventLoop(SharedResource[asyncio.AbstractEventLoop]):
    """
    A class that manages a shared asyncio event loop.
    """

    def __init__(self):
        super().__init__()
        self._close_signal_handlers: list[Callable[[], None]] = []

    def _enter(self) -> asyncio.AbstractEventLoop:
        loop = asyncio.new_event_loop()

        def worker():
            asyncio.set_event_loop(loop)
            loop.run_forever()

            not_finished_tasks = [t for t in asyncio.all_tasks(loop) if not t.done()]
            # Tasks should be given time to run even if they are cancelled so they can do cleanup
            # https://xinhuang.github.io/posts/2017-07-31-common-mistakes-using-python3-asyncio.html
            for task in not_finished_tasks:
                with suppress(CancelledError):
                    loop.run_until_complete(task)

            loop.close()

        self._loop = loop
        self._thread = threading.Thread(target=worker)
        self._thread.start()

        return loop

    def add_close_handler(self, handler: Callable[[], None]) -> None:
        self._close_signal_handlers.append(handler)

    def remove_close_handler(self, handler: Callable[[], None]):
        if handler in self._close_signal_handlers:
            self._close_signal_handlers.remove(handler)

    def signal_emit_close(self):
        for handler in self._close_signal_handlers:
            handler()

    def _exit(self):
        self._close_signal_handlers.clear()
        self._loop.call_soon_threadsafe(lambda: self._loop.stop())
        self._thread.join()

    def get_running_loop(self) -> Optional[asyncio.AbstractEventLoop]:
        return self._resource


shared_event_loop = _SharedEventLoop()


def handle_shutdown_event(*_args) -> None:
    """
    Handles the shutdown event.
    """
    shared_event_loop.signal_emit_close()



from pathlib import Path
from typing import Optional, Union

from genai._types import EnumLike
from genai._utils.general import to_enum, to_enum_optional
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai._utils.validators import assert_is_not_empty_string
from genai.schema import (
    FileCreateResponse,
    FileIdRetrieveResponse,
    FileListSortBy,
    FilePurpose,
    FileRetrieveResponse,
    SortDirection,
)
from genai.schema._api import (
    _FileCreateParametersQuery,
    _FileCreateRequest,
    _FileIdContentRetrieveParametersQuery,
    _FileIdDeleteParametersQuery,
    _FileIdRetrieveParametersQuery,
    _FileRetrieveParametersQuery,
)
from genai.schema._endpoints import (
    FileCreateEndpoint,
    FileIdContentRetrieveEndpoint,
    FileIdDeleteEndpoint,
    FileIdRetrieveEndpoint,
    FileRetrieveEndpoint,
)

__all__ = ["FileService"]


class FileService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=FileCreateEndpoint)
    def create(
        self,
        file_path: Union[str, Path],
        purpose: EnumLike[FilePurpose],
    ) -> FileCreateResponse:
        """
        Args:
            file_path: The path to the local file that will be uploaded.
            purpose: The purpose of the file to be created.

        Raises:
            ValueError: If the file does not exist or if the file format is not supported.
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        file_path = Path(file_path)
        if not file_path.is_file():
            raise ValueError(f"File {file_path} does not exist!")

        if not file_path.name.endswith(".json") and not file_path.name.endswith(".jsonl"):
            raise ValueError(f"File {file_path} must be in 'json' or jsonl 'format'!")

        with file_path.open("rb") as file_read_stream:
            request_body = _FileCreateRequest(purpose=to_enum(FilePurpose, purpose), file=b"").model_dump(
                exclude="file"
            )
            self._log_method_execution("File Create", **request_body)

            with self._get_http_client() as client:
                metadata = get_service_action_metadata(self.create)
                response = client.post(
                    url=self._get_endpoint(metadata.endpoint),
                    params=_FileCreateParametersQuery().model_dump(),
                    files={"file": (file_path.name, file_read_stream)},
                    data=request_body,
                )
                return FileCreateResponse(**response.json())

    @set_service_action_metadata(endpoint=FileIdContentRetrieveEndpoint)
    def read(
        self,
        id: str,
    ) -> str:
        """
        Args:
            id (str): The ID of the file to be read.

        Returns:
            str: The content of the file as a UTF-8 decoded string.

        Raises:
            ValueError: If the provided `id` is an empty string.
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.

        """
        assert_is_not_empty_string(id)
        self._log_method_execution("File Read", id=id)
        metadata = get_service_action_metadata(self.read)

        with self._get_http_client() as client:
            response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_FileIdContentRetrieveParametersQuery().model_dump(),
            )
            return response.content.decode("utf-8")

    @set_service_action_metadata(endpoint=FileIdRetrieveEndpoint)
    def retrieve(
        self,
        id: str,
    ) -> FileIdRetrieveResponse:
        """
        Args:
            id (str): The ID of the file to retrieve.

        Returns:
            FileIdRetrieveResponse: The response object containing the retrieved file information.

        Raises:
            ValueError: If the ID is an empty string.
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.

        """
        assert_is_not_empty_string(id)
        self._log_method_execution("File Retrieve", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.retrieve)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_FileIdRetrieveParametersQuery().model_dump(),
            )
            return FileIdRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=FileRetrieveEndpoint)
    def list(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_by: Optional[EnumLike[FileListSortBy]] = None,
        direction: Optional[EnumLike[SortDirection]] = None,
        search: Optional[str] = None,
        purpose: Optional[EnumLike[FilePurpose]] = None,
        format_id: Optional[int] = None,
    ) -> FileRetrieveResponse:
        """
        Args:
            limit: The maximum number of files to retrieve. Defaults to None.
            offset: The number of files to skip before starting retrieval. Defaults to None.
            sort_by: The field to sort the files by.
            direction: The sort direction. Can be either "asc" or "desc". Defaults to None.
            search: The search string to filter files by. Defaults to None.
            purpose: The purpose of the files. Can be one of the values from the FilePurpose enum. Defaults to None.
            format_id: The ID of the file format. Defaults to None.

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        request_params = _FileRetrieveParametersQuery(
            limit=limit,
            offset=offset,
            sort_by=to_enum_optional(sort_by, FileListSortBy),
            direction=to_enum_optional(direction, SortDirection),
            search=search,
            purpose=to_enum_optional(purpose, FilePurpose),
            format_id=format_id,
        ).model_dump()
        self._log_method_execution("File List", **request_params)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.list)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint),
                params=request_params,
            )
            return FileRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=FileIdDeleteEndpoint)
    def delete(self, id: str) -> None:
        """
        Deletes a file with the given ID.

        Args:
            id: The ID of the file to be deleted.

        Raises:
            ValueError: If the ID is an empty string.
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        assert_is_not_empty_string(id)
        self._log_method_execution("File Delete", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.delete)
            client.delete(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_FileIdDeleteParametersQuery().model_dump(),
            )



import typing
from typing import Any, Optional

import httpx
from httpx._models import Headers, Response
from httpx._types import HeaderTypes, QueryParamTypes
from httpx_sse import SSEError, connect_sse

from genai._utils.general import merge_objects
from genai._utils.responses import BaseErrorResponse
from genai.exceptions import ApiResponseException


class HttpxClient(httpx.Client):
    """
    This class represents an HTTP client based on the `httpx.Client` class with some extra methods.

    Note:
        use HttpxClient instead of httpx.Client across the project
    """

    def post_stream(
        self,
        url: str,
        *,
        headers: Optional[HeaderTypes] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
    ):
        with connect_sse(
            client=self,
            method="POST",
            url=url,
            headers=merge_objects(self.headers, Headers(headers)),
            json=json,
            params=params,
        ) as event_source:
            try:
                for sse in event_source.iter_sse():
                    if sse.event == "error":
                        if sse.data.startswith("{") and sse.data.endswith("}"):
                            raise ApiResponseException(response=BaseErrorResponse(**sse.json()))
                        else:
                            raise ValueError(f"Invalid server response during streaming!\nRetrieved data: {sse.data}")

                    yield sse.data
            except SSEError as e:
                response: Response = event_source.response
                if "application/json" in response.headers["content-type"]:
                    response.read()
                    raise ApiResponseException.from_http_response(
                        message="Invalid data chunk retrieved during streaming.", response=response
                    ) from None
                else:
                    raise e


class AsyncHttpxClient(httpx.AsyncClient):
    """
    This class represents an asynchronous HTTP client based on the `httpx.AsyncClient` class.

    Note: use AsyncHttpxClient instead of httpx.AsyncClient across the project
    """


class ReusableAsyncHttpxClient(AsyncHttpxClient):
    """
    This class extends the `AsyncHttpxClient` class and allows for reusing the client instance by maintaining a
    reference count. It provides the ability to close the client when the reference count reaches zero, as well as
    entering and exiting context for managing the reference count.
    """

    def __init__(self, *args, on_before_close_callback: typing.Callable, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._ref_count = 0
        self._on_before_close_callback = on_before_close_callback

    async def aclose(self) -> None:
        self._ref_count -= 1
        if self._ref_count == 0:
            self._ref_count = 0
            self._on_before_close_callback()
            await super().aclose()

    async def __aenter__(self):
        if self._ref_count > 0:
            self._ref_count += 1
            return self

        if self._ref_count == 0:
            self._ref_count = 1
            await super().__aenter__()

        return self

    async def __aexit__(self, *args):
        self._ref_count -= 1

        if self._ref_count == 0:
            self._on_before_close_callback()
            await super().__aexit__(*args)
            self._ref_count = 0



from queue import Queue
from typing import Generic, Optional, TypeVar

Order = int
T = TypeVar("T", bound=tuple[Order, ...])


class OrderedQueue(Queue, Generic[T]):
    """
    Special queue which ensures that items are processed in linear fashion (1, 2, 3, ...)
    The order parameter is specified as first item of the tuple.
    """

    def __init__(self, maxsize: int = 0, start: int = 0):
        super().__init__(maxsize)
        self.queue: dict[int, T] = {}
        self._expected_idx = start
        self._current_idx = start

    def put(self, item: T, block: bool = True, timeout: Optional[float] = None) -> None:
        if item[0] is None:
            return self.put((self._expected_idx, *item[1:]), block, timeout)

        super().put(item, block, timeout)

    def flush(self) -> None:
        """
        Flushes all the items in the queue (does not wait for the consumer).
        """
        while not self.empty():
            self.queue.popitem()
            self.task_done()

    def get(self, block=True, timeout=None) -> T:
        """
        Get an item from the queue (waits until someone adds the item to the queue with given index)

        Args:
            block (bool): Indicates whether the method should block until an item is available in the queue.
            timeout (Optional[float]): The maximum amount of time (in seconds) that the method should wait.

        Returns:
            T: The item retrieved from the queue.

        """
        with self.not_empty:
            target = self._current_idx
            self._current_idx += 1

            item = self.not_empty.wait_for(
                lambda: self._get() if target == self._expected_idx and target in self.queue else None
            )
            assert item

            self._expected_idx += 1
            self.not_full.notify()
            if not self._qsize():
                self.not_empty.notify_all()

            return item

    def _get(self) -> T:
        return self.queue.pop(self._expected_idx)

    def _put(self, item: T):
        idx: Order = item[0]
        if idx in self.queue:
            raise ValueError(f"Entry with key '{idx}' already exists!")
        self.queue[idx] = item

    def _init(self, maxsize: int):
        self.queue = {}



import asyncio
import logging
from functools import partial
from typing import Any, Optional, Sequence

from genai import Client
from genai._types import EnumLike
from genai.extensions._common.utils import (
    _prepare_chat_generation_request,
    create_generation_info_from_response,
)
from genai.schema import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    ModerationParameters,
    PromptTemplateData,
    SystemMessage,
    TextGenerationParameters,
    TrimMethod,
)

logger = logging.getLogger(__name__)


try:
    from llama_index.callbacks import CallbackManager
    from llama_index.llms.base import (
        ChatMessage,
        ChatResponse,
        ChatResponseAsyncGen,
        ChatResponseGen,
        CompletionResponse,
        CompletionResponseAsyncGen,
        CompletionResponseGen,
        LLMMetadata,
        llm_chat_callback,
        llm_completion_callback,
    )
    from llama_index.llms.llm import LLM
    from llama_index.llms.types import MessageRole


except ImportError:
    raise ImportError("Could not import llamaindex: Please install ibm-generative-ai[llama-index] extension.")  # noqa: B904


def to_genai_message(message: ChatMessage) -> BaseMessage:
    content = message.content or ""
    if message.role == MessageRole.USER:
        return HumanMessage(content=content)
    elif message.role == MessageRole.SYSTEM:
        return SystemMessage(content=content)
    elif message.role == MessageRole.ASSISTANT:
        return AIMessage(content=content)
    else:
        raise ValueError(f"Got unknown message type {message}")


def to_genai_messages(messages: Sequence[ChatMessage]) -> list[BaseMessage]:
    return [to_genai_message(msg) for msg in messages]


class IBMGenAILlamaIndex(LLM):
    client: Client

    # Generation related parameters
    model_id: str
    prompt_id: Optional[str] = None
    parameters: Optional[TextGenerationParameters] = None
    moderations: Optional[ModerationParameters] = None
    data: Optional[PromptTemplateData] = None

    # Chat (only) related parameters
    conversation_id: Optional[str] = None
    parent_id: Optional[str] = None
    prompt_template_id: Optional[str] = None
    trim_method: Optional[EnumLike[TrimMethod]] = None
    use_conversation_parameters: Optional[bool] = None

    def __init__(
        self, *, client: Client, model_id: str, callback_manager: Optional[CallbackManager] = None, **kwargs: Any
    ):
        super().__init__(
            client=client, callback_manager=callback_manager or CallbackManager(), model_id=model_id, **kwargs
        )

    @classmethod
    def class_name(cls) -> str:
        return "ibmgenai_llm"

    @property
    def _common_indentifying_params(self) -> dict[str, Any]:
        return {
            "model_id": self.model_id,
            "prompt_id": self.prompt_id,
            "parameters": self.parameters,
            "moderations": self.moderations,
        }

    @property
    def _identifying_params(self) -> dict[str, Any]:
        return {
            **self._common_indentifying_params,
            "data": self.data,
        }

    @property
    def _identifying_chat_params(self) -> dict[str, Any]:
        return {
            **self._common_indentifying_params,
            "conversation_id": self.conversation_id,
            "parent_id": self.parent_id,
            "prompt_template_id": self.prompt_template_id,
            "trim_method": self.trim_method,
            "use_conversation_parameters": self.use_conversation_parameters,
        }

    def _prepare_request(self, source: dict[str, Any]):
        def prepare(**kwargs: Any):
            updated = {k: kwargs.pop(k, v) for k, v in source.items()}
            return _prepare_chat_generation_request(
                **kwargs,
                **updated,
            )

        return prepare

    @property
    def metadata(self) -> LLMMetadata:
        model = self.client.model.retrieve(self.model_id).result
        assert model.token_limits[0].token_limit
        context_window = int(model.token_limits[0].token_limit)
        return LLMMetadata(
            context_window=context_window,
            is_chat_model=True,
            is_function_calling_model=False,
            model_name=model.name or self.model_id,
        )

    @llm_completion_callback()
    def complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponse:
        if not formatted:
            prompt = self.completion_to_prompt(prompt)

        response = list(
            self.client.text.generation.create(
                **self._prepare_request(self._identifying_params)(inputs=[prompt], **kwargs)
            )
        )[0]
        result = response.results[0]
        generation_info = create_generation_info_from_response(response, result=result)

        return CompletionResponse(text=result.generated_text or "", additional_kwargs=generation_info)

    @llm_completion_callback()
    def stream_complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponseGen:
        if not formatted:
            prompt = self.completion_to_prompt(prompt)

        text = ""
        for response in self.client.text.generation.create_stream(
            **self._prepare_request(self._identifying_params)(input=prompt, **kwargs)
        ):
            for result in response.results or []:
                generated_text = result.generated_text or ""
                generation_info = create_generation_info_from_response(response, result=result)

                text += generated_text
                yield CompletionResponse(text=text, delta=generated_text, additional_kwargs=generation_info)

    @llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        genai_messages = to_genai_messages(messages)
        response = self.client.text.chat.create(
            **self._prepare_request(self._identifying_chat_params)(messages=genai_messages, **kwargs)
        )
        result = response.results[0]
        generation_info = create_generation_info_from_response(response, result=result)

        return ChatResponse(
            message=ChatMessage(
                role=MessageRole.ASSISTANT,
                content=result.generated_text or "",
            ),
            additional_kwargs=generation_info,
        )

    @llm_chat_callback()
    def stream_chat(self, messages: Sequence[ChatMessage], formatted: bool = False, **kwargs: Any) -> ChatResponseGen:
        text = ""

        for response in self.client.text.chat.create_stream(
            **self._prepare_request(self._identifying_chat_params)(messages=to_genai_messages(messages), **kwargs)
        ):
            if response.moderation:
                generation_info = create_generation_info_from_response(response, result=response.moderation)
                message = ChatMessage(role=MessageRole.ASSISTANT, content=text)
                yield ChatResponse(message=message, delta="", additional_kwargs=generation_info)

            for result in response.results or []:
                generated_text = result.generated_text or ""
                generation_info = create_generation_info_from_response(response, result=result)
                text += generated_text
                message = ChatMessage(role=MessageRole.ASSISTANT, content=text)
                yield ChatResponse(message=message, delta=generated_text, additional_kwargs=generation_info)

    @llm_completion_callback()
    async def acomplete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponse:
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(self.complete, prompt, formatted, **kwargs)
        )

    @llm_completion_callback()
    async def astream_complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponseAsyncGen:
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(self.stream_complete, prompt, formatted, **kwargs)
        )

    @llm_chat_callback()
    async def achat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        return await asyncio.get_running_loop().run_in_executor(None, partial(self.chat, messages, **kwargs))

    @llm_chat_callback()
    async def astream_chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponseAsyncGen:
        return await asyncio.get_running_loop().run_in_executor(None, partial(self.stream_chat, messages, **kwargs))



from typing import Optional

from pydantic import BaseModel

from genai._utils.api_client import ApiClient
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
)
from genai.text.chat.chat_generation_service import ChatService as _ChatService
from genai.text.embedding.embedding_service import EmbeddingService as _EmbeddingService
from genai.text.generation.generation_service import GenerationService as _GenerationService
from genai.text.moderation.moderation_service import ModerationService as _ModerationService
from genai.text.tokenization.tokenization_service import TokenizationService as _TokenizationService

__all__ = ["TextService", "BaseServices"]


class BaseServices(BaseModel):
    """Appropriate services used by the Text Service"""

    GenerationService: type[_GenerationService] = _GenerationService
    TokenizationService: type[_TokenizationService] = _TokenizationService
    ChatService: type[_ChatService] = _ChatService
    ModerationService: type[_ModerationService] = _ModerationService
    EmbeddingService: type[_EmbeddingService] = _EmbeddingService


class TextService(BaseService[BaseServiceConfig, BaseServices]):
    """
    Class providing access to various text-related services.
    """

    Services = BaseServices

    def __init__(self, *, api_client: ApiClient, services: Optional[BaseServices] = None, **kwargs):
        super().__init__(api_client=api_client, **kwargs)

        if not services:
            services = self.Services()

        self.moderation = services.ModerationService(api_client=api_client)
        self.generation = services.GenerationService(api_client=api_client)
        self.tokenization = services.TokenizationService(api_client=api_client)
        self.chat = services.ChatService(api_client=api_client)
        self.embedding = services.EmbeddingService(api_client=api_client)



# generated by datamodel-codegen:
#   filename:  2024-01-30_openapi_schema

from __future__ import annotations

from datetime import date
from enum import Enum
from typing import Any, Literal, Optional, Union

from pydantic import AwareDatetime, Field, RootModel

from genai._types import ApiBaseModel


class ApiKeyResult(ApiBaseModel):
    created_at: AwareDatetime
    generated_at: AwareDatetime
    last_used_at: Optional[AwareDatetime] = None
    value: str


class BaseErrorExtension(ApiBaseModel):
    code: str
    state: Optional[dict[str, Any]] = None


class BaseErrorResponse(ApiBaseModel):
    error: str
    extensions: BaseErrorExtension
    message: str
    status_code: int


class ChatRole(str, Enum):
    USER = "user"
    SYSTEM = "system"
    ASSISTANT = "assistant"


class ConcurrencyLimit(ApiBaseModel):
    limit: int
    remaining: int


class DecodingMethod(str, Enum):
    GREEDY = "greedy"
    SAMPLE = "sample"


class FileFormat(ApiBaseModel):
    id: int
    name: str


class FileListSortBy(str, Enum):
    NAME = "name"
    CREATED_AT = "created_at"


class FilePurpose(str, Enum):
    TUNE = "tune"
    TEMPLATE = "template"
    TUNE_IMPORT = "tune_import"
    EXTRACTION = "extraction"


class GeneratedToken(ApiBaseModel):
    logprob: Optional[Union[float, str]] = None
    text: Optional[str] = None


class HAPOptions(ApiBaseModel):
    send_tokens: Optional[bool] = False
    threshold: Optional[float] = Field(0.75, gt=0.0, lt=1.0)


class ImplicitHateOptions(ApiBaseModel):
    send_tokens: Optional[bool] = False
    threshold: Optional[float] = Field(0.75, gt=0.0, lt=1.0)


class Extensions1(BaseErrorExtension):
    code: Literal["INTERNAL_SERVER_ERROR"] = "INTERNAL_SERVER_ERROR"


class InternalServerErrorResponse(BaseErrorResponse):
    extensions: Extensions1
    status_code: Literal[500] = 500


class LengthPenalty(ApiBaseModel):
    decay_factor: Optional[float] = Field(None, gt=1.0, title="Decay factor")
    """
    Represents the factor of exponential decay and must be > 1.0. Larger values correspond to more aggressive decay.
    """
    start_index: Optional[int] = Field(None, ge=1, title="Start index")
    """
    A number of generated tokens after which this should take effect.
    """


class ModelFamily(ApiBaseModel):
    description: Optional[str] = None
    id: int
    name: str
    prompt_example: Optional[str] = None
    short_description: Optional[str] = None
    system_prompt: Optional[str] = None


class ModelTokenLimits(ApiBaseModel):
    beam_width: int
    token_limit: int


class ModerationHAP(ApiBaseModel):
    input: Optional[bool] = True
    """
    Detects HAP (hateful, abusive, or profane language). Please see documentation for more info (API Reference -> Moderations -> HAP).
    """
    output: Optional[bool] = None
    send_tokens: Optional[bool] = False
    threshold: Optional[float] = Field(0.75, ge=0.01, le=0.99, multiple_of=0.01)
    """
    The higher the number, the more confidence that the sentence contains HAP. The threshold allows you to modify how much confidence is needed for the sentence to be flagged as containing HAP.
    """


class ModerationImplicitHate(ApiBaseModel):
    input: Optional[bool] = True
    """
    Detects implicit hate.
    """
    output: Optional[bool] = None
    send_tokens: Optional[bool] = False
    threshold: Optional[float] = Field(0.75, ge=0.01, le=0.99, multiple_of=0.01)
    """
    The higher the number, the more confidence that the sentence contains implicit hate. The threshold allows you to modify how much confidence is needed for the sentence to be flagged as containing implicit hate.
    """


class ModerationPosition(ApiBaseModel):
    end: int
    start: int


class ModerationStigma(ApiBaseModel):
    input: Optional[bool] = True
    """
    Detects Stigma.
    """
    output: Optional[bool] = None
    send_tokens: Optional[bool] = False
    threshold: Optional[float] = Field(0.75, ge=0.01, le=0.99, multiple_of=0.01)
    """
    The higher the number, the more confidence that the sentence contains Stigma. The threshold allows you to modify how much confidence is needed for the sentence to be flagged as containing Stigma.
    """


class ModerationTokens(ApiBaseModel):
    index: Optional[int] = None
    score: Optional[float] = None
    token: Optional[str] = None


class Extensions2(BaseErrorExtension):
    code: Literal["NOT_FOUND"] = "NOT_FOUND"


class NotFoundResponse(BaseErrorResponse):
    extensions: Extensions2
    status_code: Literal[404] = 404


class PromptResultAuthor(ApiBaseModel):
    first_name: Optional[str] = None
    id: Optional[int] = None
    last_name: Optional[str] = None


class PromptResultTask(ApiBaseModel):
    icon: Optional[str] = None
    id: Optional[str] = None
    name: Optional[str] = None


class PromptRetrieveRequestParamsSource(str, Enum):
    USER = "user"
    EXAMPLE = "example"
    COMMUNITY = "community"


class PromptTemplateData(ApiBaseModel):
    example_file_ids: Optional[list[str]] = Field(None, max_length=5, min_length=0)


class PromptType(str, Enum):
    PRIVATE = "private"
    PUBLIC = "public"
    COMMUNITY = "community"
    EXAMPLE = "example"


class RequestApiVersion(str, Enum):
    V0 = "v0"
    V1 = "v1"
    V2 = "v2"


class RequestEndpoint(str, Enum):
    GENERATE = "generate"
    COMPARE = "compare"
    CHAT = "chat"


class RequestOrigin(str, Enum):
    API = "api"
    UI = "ui"


class RequestResultVersion(ApiBaseModel):
    api: Optional[str] = None
    date_: Optional[date] = Field(None, alias="date")


class RequestStatus(str, Enum):
    SUCCESS = "success"
    ERROR = "error"


class SortDirection(str, Enum):
    ASC = "asc"
    DESC = "desc"


class StigmaOptions(ApiBaseModel):
    send_tokens: Optional[bool] = False
    threshold: Optional[float] = Field(0.75, gt=0.0, lt=1.0)


class StopReason(str, Enum):
    NOT_FINISHED = "not_finished"
    MAX_TOKENS = "max_tokens"
    EOS_TOKEN = "eos_token"
    CANCELLED = "cancelled"
    TIME_LIMIT = "time_limit"
    STOP_SEQUENCE = "stop_sequence"
    TOKEN_LIMIT = "token_limit"
    ERROR = "error"


class StorageProviderLocation(str, Enum):
    US_SOUTH = "us-south"
    US_EAST = "us-east"


class SystemPromptAuthor(ApiBaseModel):
    first_name: Optional[str] = None
    id: int
    last_name: Optional[str] = None


class SystemPromptType(str, Enum):
    PRIVATE = "private"
    SYSTEM = "system"


class Tasks(ApiBaseModel):
    categorization: bool
    csv_example: Optional[str] = None
    file_format_id: Optional[int] = None
    id: str
    json_example: Optional[str] = None
    jsonl_example: Optional[str] = None
    name: str
    tune: bool
    verbalizer: Optional[str] = None


class TextEmbeddingLimit(ApiBaseModel):
    concurrency: ConcurrencyLimit


class TextEmbeddingParameters(ApiBaseModel):
    truncate_input_tokens: Optional[bool] = None


class TextGenerationComparisonCreateResultsParameters(ApiBaseModel):
    length_penalty: Optional[dict[str, Any]] = None
    model_id: Optional[str] = None
    repetition_penalty: Optional[float] = None
    temperature: Optional[float] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    typical_p: Optional[float] = None


class TextGenerationFeedbackCategory(str, Enum):
    NO_PROBLEM = "no_problem"
    HAP = "hap"
    PII = "pii"
    SOCIAL_BIAS = "social_bias"
    NOT_HONEST_OR_TRUTHFUL = "not_honest_or_truthful"
    TABOO_TOPICS = "taboo_topics"
    OTHER = "other"


class TextGenerationFeedbackResult(ApiBaseModel):
    api_request: str
    categories: list[str]
    comment: Optional[str] = None
    created_at: AwareDatetime
    id: int
    updated_at: AwareDatetime


class TextGenerationLimit(ApiBaseModel):
    concurrency: ConcurrencyLimit


class TextGenerationReturnOptions(ApiBaseModel):
    generated_tokens: Optional[bool] = Field(False, title="Generated Tokens")
    """
    Include list of individual generated tokens
    """
    input_parameters: Optional[bool] = None
    input_text: Optional[bool] = Field(False, title="Input text")
    """
    Include input text
    """
    input_tokens: Optional[bool] = Field(False, title="Input Tokens")
    """
    Include list of input tokens
    """
    token_logprobs: Optional[bool] = Field(False, title="Token logprobs")
    """
    Include logprob for each returned token
    """
    token_ranks: Optional[bool] = Field(False, title="Token ranks")
    """
    Include rank of each returned token
    """
    top_n_tokens: Optional[int] = Field(None, ge=0, le=5, title="Top N tokens")
    """
    Include top n candidate tokens at the position of each returned token
    """


class TextModeration(ApiBaseModel):
    flagged: bool
    position: ModerationPosition
    score: float
    success: bool
    tokens: Optional[list[ModerationTokens]] = None


class TextTokenizationCreateResults(ApiBaseModel):
    input_text: Optional[str] = None
    token_count: int
    tokens: Optional[list[str]] = None


class TextTokenizationReturnOptions(ApiBaseModel):
    input_text: Optional[bool] = None
    tokens: Optional[bool] = None


class Extensions3(BaseErrorExtension):
    code: Literal["TOO_MANY_REQUESTS"] = "TOO_MANY_REQUESTS"


class TooManyRequestsResponse(BaseErrorResponse):
    extensions: Extensions3
    status_code: Literal[429] = 429


class TrimMethod(str, Enum):
    FLOATING_WINDOW = "floating_window"
    NONE = "none"


class TuneAssetType(str, Enum):
    VECTORS = "vectors"
    LOGS = "logs"
    EXPORT = "export"


class TuneListSortBy(str, Enum):
    STATUS = "status"
    CREATED_AT = "created_at"
    NAME = "name"
    ID = "id"
    MODEL = "model"


class TuneParameters(ApiBaseModel):
    accumulate_steps: Optional[int] = None
    batch_size: Optional[int] = None
    learning_rate: Optional[float] = None
    max_input_tokens: Optional[int] = None
    max_output_tokens: Optional[int] = None
    num_epochs: Optional[int] = None
    num_virtual_tokens: Optional[int] = None
    verbalizer: Optional[str] = None


class TuneResultDatapointLoss(ApiBaseModel):
    epoch: int
    step: Optional[int] = None
    value: float


class TuneResultFiles(ApiBaseModel):
    created_at: Optional[AwareDatetime] = None
    file_name: str
    id: str


class TuneStatus(str, Enum):
    INITIALIZING = "initializing"
    NOT_STARTED = "not_started"
    PENDING = "pending"
    HALTED = "halted"
    RUNNING = "running"
    QUEUED = "queued"
    COMPLETED = "completed"
    FAILED = "failed"


class TunesResultDatapointLoss(ApiBaseModel):
    data: TuneResultDatapointLoss
    timestamp: AwareDatetime


class TuningType(str, Enum):
    PROMPT_TUNING = "prompt_tuning"
    MULTITASK_PROMPT_TUNING = "multitask_prompt_tuning"


class TuningTypeRetrieveResults(ApiBaseModel):
    id: Optional[str] = None
    model_ids: Optional[list[str]] = None
    name: Optional[str] = None
    schema_: Optional[dict[str, Any]] = Field(None, alias="schema")
    """
    JSON Schema
    """


class Extensions4(BaseErrorExtension):
    code: Literal["AUTH_ERROR"] = "AUTH_ERROR"


class UnauthorizedResponse(BaseErrorResponse):
    extensions: Extensions4
    status_code: Literal[401] = 401


class Extensions5(BaseErrorExtension):
    code: Literal["SERVICE_UNAVAILABLE"] = "SERVICE_UNAVAILABLE"


class UnavailableResponse(BaseErrorResponse):
    extensions: Extensions5
    status_code: Literal[503] = 503


class UserCreateResultApiKey(ApiBaseModel):
    created_at: str
    generated_at: str
    last_used_at: Optional[str] = None
    value: str


class _ApiKeyRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class ApiKeyRetrieveResponse(ApiBaseModel):
    result: Optional[ApiKeyResult] = None


class _ApiKeyRegenerateCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class ApiKeyRegenerateCreateResponse(ApiBaseModel):
    result: Optional[ApiKeyResult] = None


class _FileRetrieveParametersQuery(ApiBaseModel):
    limit: Optional[int] = Field(100, ge=1, le=100)
    offset: Optional[int] = Field(0, ge=0)
    sort_by: Optional[FileListSortBy] = None
    direction: Optional[SortDirection] = None
    search: Optional[str] = None
    purpose: Optional[FilePurpose] = None
    format_id: Optional[int] = None
    version: Literal["2023-12-15"] = "2023-12-15"


class _FileCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-12-15"] = "2023-12-15"


class _FileCreateRequest(ApiBaseModel):
    file: bytes
    purpose: FilePurpose


class _FileIdDeleteParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _FileIdRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-12-15"] = "2023-12-15"


class _FileIdContentRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _ModelRetrieveParametersQuery(ApiBaseModel):
    limit: Optional[int] = Field(100, ge=1, le=100)
    offset: Optional[int] = Field(0, ge=0)
    version: Literal["2023-11-22"] = "2023-11-22"


class _ModelIdRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2024-01-30"] = "2024-01-30"


class _PromptRetrieveParametersQuery(ApiBaseModel):
    limit: Optional[int] = Field(100, ge=1, le=100)
    offset: Optional[int] = Field(0, ge=0)
    search: Optional[str] = None
    task_id: Optional[Union[str, list[str]]] = None
    model_id: Optional[Union[str, list[str]]] = None
    source: Optional[Union[PromptRetrieveRequestParamsSource, list[PromptRetrieveRequestParamsSource]]] = None
    version: Literal["2024-01-10"] = "2024-01-10"


class _PromptCreateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _PromptIdDeleteParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _PromptIdRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _PromptIdPatchParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _PromptIdPatchRequest(ApiBaseModel):
    type: Optional[PromptType] = None


class _PromptIdUpdateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _RequestRetrieveParametersQuery(ApiBaseModel):
    limit: Optional[int] = Field(100, ge=1, le=100)
    offset: Optional[int] = Field(0, ge=0)
    status: Optional[RequestStatus] = None
    origin: Optional[RequestOrigin] = None
    before: Optional[AwareDatetime] = None
    after: Optional[AwareDatetime] = None
    endpoint: Optional[Union[RequestEndpoint, list[RequestEndpoint]]] = None
    api: Optional[RequestApiVersion] = None
    date_: Optional[date] = Field(None, alias="date")
    version: Literal["2023-11-22"] = "2023-11-22"


class _RequestChatConversationIdDeleteParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _RequestChatConversationIdRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _RequestIdDeleteParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _SystemPromptRetrieveParametersQuery(ApiBaseModel):
    limit: Optional[int] = Field(100, ge=1, le=100)
    offset: Optional[int] = Field(0, ge=0)
    version: Literal["2023-11-22"] = "2023-11-22"


class _SystemPromptCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _SystemPromptCreateRequest(ApiBaseModel):
    content: str
    name: str


class _SystemPromptIdDeleteParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _SystemPromptIdRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _SystemPromptIdUpdateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _SystemPromptIdUpdateRequest(ApiBaseModel):
    content: str
    name: str


class _TaskRetrieveParametersQuery(ApiBaseModel):
    tune: Optional[bool] = True
    version: Literal["2023-11-22"] = "2023-11-22"


class TaskRetrieveResponse(ApiBaseModel):
    results: list[Tasks]


class _TextChatCreateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _TextChatOutputCreateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class TextChatOutputCreateResponse(ApiBaseModel):
    result: str


class _TextChatStreamCreateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _TextEmbeddingCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class Input(RootModel[list[Any]]):
    root: list[Any] = Field(..., max_length=20)


class _TextEmbeddingCreateRequest(ApiBaseModel):
    input: Union[str, Input]
    model_id: str
    parameters: Optional[TextEmbeddingParameters] = None


class TextEmbeddingCreateResponse(ApiBaseModel):
    results: list[list[float]]


class _TextEmbeddingLimitRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class TextEmbeddingLimitRetrieveResponse(ApiBaseModel):
    result: TextEmbeddingLimit


class _TextExtractionLimitRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TextGenerationCreateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _TextGenerationComparisonCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TextGenerationLimitRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class TextGenerationLimitRetrieveResponse(ApiBaseModel):
    result: TextGenerationLimit


class _TextGenerationOutputCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class TextGenerationOutputCreateResponse(ApiBaseModel):
    results: list[str]


class _TextGenerationIdFeedbackRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class TextGenerationIdFeedbackRetrieveResponse(ApiBaseModel):
    result: TextGenerationFeedbackResult


class _TextGenerationIdFeedbackCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TextGenerationIdFeedbackCreateRequest(ApiBaseModel):
    categories: Optional[list[TextGenerationFeedbackCategory]] = Field(None, min_length=1)
    comment: Optional[str] = None


class TextGenerationIdFeedbackCreateResponse(ApiBaseModel):
    result: TextGenerationFeedbackResult


class _TextGenerationIdFeedbackUpdateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TextGenerationIdFeedbackUpdateRequest(ApiBaseModel):
    categories: Optional[list[TextGenerationFeedbackCategory]] = Field(None, min_length=1)
    comment: Optional[str] = None


class TextGenerationIdFeedbackUpdateResponse(ApiBaseModel):
    result: TextGenerationFeedbackResult


class _TextGenerationStreamCreateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class _TextModerationCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TextModerationCreateRequest(ApiBaseModel):
    hap: Optional[HAPOptions] = None
    implicit_hate: Optional[ImplicitHateOptions] = None
    input: str = Field(..., max_length=20480)
    stigma: Optional[StigmaOptions] = None


class _TextTokenizationCreateParametersQuery(ApiBaseModel):
    version: Literal["2024-01-10"] = "2024-01-10"


class TextTokenizationCreateResponse(ApiBaseModel):
    created_at: str
    model_id: str
    results: list[TextTokenizationCreateResults]


class _TuneRetrieveParametersQuery(ApiBaseModel):
    limit: Optional[int] = Field(100, ge=1, le=100)
    offset: Optional[int] = Field(0, ge=0)
    status: Optional[TuneStatus] = None
    search: Optional[str] = None
    sort_by: Optional[TuneListSortBy] = None
    direction: Optional[SortDirection] = None
    version: Literal["2023-11-22"] = "2023-11-22"


class _TuneCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TuneCreateRequest(ApiBaseModel):
    evaluation_file_ids: Optional[list[str]] = None
    model_id: str
    name: str
    parameters: Optional[TuneParameters] = None
    task_id: str
    training_file_ids: list[str]
    tuning_type: TuningType
    validation_file_ids: Optional[list[str]] = None


class _TuneImportCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TuneImportCreateRequest(ApiBaseModel):
    file_id: str
    name: str


class _TuneIdDeleteParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TuneIdRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TuneIdPatchParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _TuneIdPatchRequest(ApiBaseModel):
    name: Optional[str] = None
    preferred: Optional[bool] = None


class _TuneIdContentTypeRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-12-15"] = "2023-12-15"


class _TuningTypeRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2024-01-30"] = "2024-01-30"


class TuningTypeRetrieveResponse(ApiBaseModel):
    results: list[TuningTypeRetrieveResults]


class _UserDeleteParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _UserRetrieveParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _UserPatchParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _UserPatchRequest(ApiBaseModel):
    data_usage_consent: Optional[bool] = None
    tou_accepted: Optional[bool] = None


class _UserCreateParametersQuery(ApiBaseModel):
    version: Literal["2023-11-22"] = "2023-11-22"


class _UserCreateRequest(ApiBaseModel):
    first_name: Optional[str] = None
    last_name: Optional[str] = None


class Extensions(BaseErrorExtension):
    code: Literal["INVALID_INPUT"] = "INVALID_INPUT"


class BadRequestResponse(BaseErrorResponse):
    extensions: Extensions
    status_code: Literal[400] = 400


class BaseMessage(ApiBaseModel):
    content: str
    file_ids: Optional[list[str]] = None
    role: ChatRole


class BaseTokens(ApiBaseModel):
    logprob: Optional[Union[float, str]] = None
    rank: Optional[int] = None
    text: Optional[str] = None
    top_tokens: Optional[list[GeneratedToken]] = None


class FileResult(ApiBaseModel):
    bytes: int
    created_at: AwareDatetime
    file_formats: Optional[list[FileFormat]] = None
    file_name: str
    id: str
    purpose: FilePurpose
    storage_provider_location: StorageProviderLocation


class ModelIdRetrieveResult(ApiBaseModel):
    description: Optional[str] = None
    developer: Optional[str] = None
    disabled: bool
    id: str
    is_live: bool
    label: str
    model_family: ModelFamily
    name: str
    preferred: bool
    size: str
    source_model_id: Optional[str] = None
    system_prompt: Optional[str] = None
    system_prompt_id: Optional[int] = None
    tags: list[str]
    tasks: list[Tasks]
    token_limits: list[ModelTokenLimits]
    warning: Optional[str] = None


class ModelRetrieveResults(ApiBaseModel):
    id: str
    is_live: bool
    label: str
    name: str
    size: str
    source_model_id: Optional[str] = None
    task_ids: list[str]
    token_limits: list[ModelTokenLimits]
    warning: Optional[str] = None


class ModerationParameters(ApiBaseModel):
    hap: Optional[Union[bool, ModerationHAP]] = None
    implicit_hate: Optional[Union[bool, ModerationImplicitHate]] = None
    stigma: Optional[Union[bool, ModerationStigma]] = None


class PromptTemplate(ApiBaseModel):
    data: PromptTemplateData
    id: Optional[str] = None
    value: Optional[str] = None


class RequestRetrieveResults(ApiBaseModel):
    created_at: AwareDatetime
    duration: int
    id: str
    request: Optional[dict[str, Any]] = None
    response: Optional[dict[str, Any]] = None
    status: RequestStatus
    version: Optional[RequestResultVersion] = None


class SystemPrompt(ApiBaseModel):
    author: Optional[SystemPromptAuthor] = None
    content: str
    created_at: AwareDatetime
    id: int
    name: str
    type: SystemPromptType


class TextCreateResponseModeration(ApiBaseModel):
    hap: Optional[list[TextModeration]] = None
    implicit_hate: Optional[list[TextModeration]] = None
    stigma: Optional[list[TextModeration]] = None


class TextGenerationComparisonParameters(ApiBaseModel):
    length_penalty: Optional[list[dict[str, Any]]] = Field(None, max_length=10, min_length=1)
    model_id: Optional[list[str]] = Field(None, max_length=10, min_length=1)
    repetition_penalty: Optional[list[float]] = Field(None, max_length=10, min_length=1)
    temperature: Optional[list[float]] = Field(None, max_length=10, min_length=1)
    top_k: Optional[list[int]] = Field(None, max_length=10, min_length=1)
    top_p: Optional[list[float]] = Field(None, max_length=10, min_length=1)
    typical_p: Optional[list[float]] = Field(None, max_length=10, min_length=1)


class TextGenerationParameters(ApiBaseModel):
    beam_width: Optional[int] = Field(None, ge=0, le=3, title="Beam width")
    """
    At each step, or token, the algorithm keeps track of the n (off=1, 2, or 3) most probable sequences (beams) and selects the one with the highest probability. This continues until the stop sequence is met.
    """
    decoding_method: Optional[DecodingMethod] = None
    include_stop_sequence: Optional[bool] = None
    length_penalty: Optional[LengthPenalty] = None
    max_new_tokens: Optional[int] = Field(None, ge=0, le=4096, title="Max new tokens")
    """
    Define the maximum number of tokens to generate.
    """
    min_new_tokens: Optional[int] = Field(None, ge=0, title="Min new tokens")
    """
    If stop sequences are given, they are ignored until minimum tokens are generated.
    """
    random_seed: Optional[int] = Field(None, ge=1, le=4294967295, title="Random seed")
    """
    Controls the random sampling of the generated tokens when sampling is enabled. Setting the random seed to a the same number for each generation ensures experimental repeatability.
    """
    repetition_penalty: Optional[float] = Field(None, ge=1.0, le=2.0, multiple_of=0.01, title="Repetition penalty")
    """
    The parameter for repetition penalty. 1.00 means no penalty.
    """
    return_options: Optional[TextGenerationReturnOptions] = None
    stop_sequences: Optional[list[str]] = Field(
        None, examples=['[" and "]'], max_length=6, min_length=1, title="Stop sequences"
    )
    """
    Stop sequences are one or more strings which will cause the text generation to stop if/when they are produced as part of the output. Stop sequences encountered prior to the minimum number of tokens being generated will be ignored.
    """
    temperature: Optional[float] = Field(None, ge=0.0, le=2.0, multiple_of=0.01, title="Temperature")
    """
    Control the creativity of generated text. Higher values will lead to more randomly generated outputs.
    """
    time_limit: Optional[int] = Field(None, title="Time limit")
    """
    Time limit in milliseconds - if not completed within this time, generation will stop. The text generated so far will be returned along with the `TIME_LIMIT` stop reason.
    """
    top_k: Optional[int] = Field(None, ge=1, le=100, title="Top K")
    """
    Set the number of highest probability vocabulary tokens to keep for top-k-filtering. Lower values make it less likely the model will go off topic.
    """
    top_p: Optional[float] = Field(None, ge=0.0, le=1.0, multiple_of=0.01, title="Top P (nucleus sampling)")
    """
    If < 1.0, only the smallest set of most probable tokens with probabilities that add up to `top_p` or higher are used.
    """
    truncate_input_tokens: Optional[int] = Field(None, ge=0, title="Truncate input tokens")
    """
    Truncate to this many input tokens. Can be used to avoid requests failing due to input being longer than configured limits. Zero means don't truncate.
    """
    typical_p: Optional[float] = Field(None, ge=0.01, le=1.0, multiple_of=0.01, title="Typical P")
    """
    Local typicality measures how similar the conditional probability of predicting a target token next is to the expected conditional probability of predicting a random token next, given the partial text already generated. If set to float < 1, the smallest set of the most locally typical tokens with probabilities that add up to typical_p or higher are kept for generation. 1.00 means a neutral value.
    """


class TextGenerationResult(ApiBaseModel):
    generated_text: str
    generated_token_count: int
    generated_tokens: Optional[list[BaseTokens]] = None
    input_text: Optional[str] = None
    input_token_count: Optional[int] = None
    input_tokens: Optional[list[BaseTokens]] = None
    moderation: Optional[dict[str, Any]] = None
    seed: Optional[float] = None
    stop_reason: StopReason
    stop_sequence: Optional[str] = None


class TextGenerationStreamResult(ApiBaseModel):
    generated_text: str
    generated_token_count: int
    generated_tokens: Optional[list[BaseTokens]] = None
    input_text: Optional[str] = None
    input_token_count: Optional[int] = None
    input_tokens: Optional[list[BaseTokens]] = None
    seed: Optional[float] = None
    stop_reason: StopReason
    stop_sequence: Optional[str] = None


class TextTokenizationParameters(ApiBaseModel):
    return_options: Optional[TextTokenizationReturnOptions] = None


class TuneResultDatapoints(ApiBaseModel):
    loss: list[TunesResultDatapointLoss]


class UserGenerationDefault(ApiBaseModel):
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = None
    template: Optional[PromptTemplate] = None


class UserResponseResult(ApiBaseModel):
    data_usage_consent: bool
    first_name: Optional[str] = None
    generate_default: Optional[UserGenerationDefault] = None
    id: int
    last_name: Optional[str] = None
    tou_accepted: bool
    tou_accepted_at: Optional[str] = None


class FileRetrieveResponse(ApiBaseModel):
    results: list[FileResult]
    total_count: int


class FileCreateResponse(ApiBaseModel):
    result: FileResult


class FileIdRetrieveResponse(ApiBaseModel):
    result: FileResult


class ModelRetrieveResponse(ApiBaseModel):
    results: list[ModelRetrieveResults]


class ModelIdRetrieveResponse(ApiBaseModel):
    result: ModelIdRetrieveResult


class _PromptCreateRequest(ApiBaseModel):
    data: Optional[PromptTemplateData] = None
    description: Optional[str] = None
    input: Optional[str] = None
    messages: Optional[list[BaseMessage]] = None
    model_id: str
    moderations: Optional[ModerationParameters] = None
    name: str
    output: Optional[str] = Field(None, min_length=1)
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = None
    task_id: Optional[str] = None
    type: Optional[PromptType] = None


class _PromptIdUpdateRequest(ApiBaseModel):
    data: Optional[PromptTemplateData] = None
    description: Optional[str] = None
    input: Optional[str] = None
    messages: Optional[list[BaseMessage]] = None
    model_id: str
    moderations: Optional[ModerationParameters] = None
    name: str
    output: Optional[str] = Field(None, min_length=1)
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = None
    task_id: Optional[str] = None
    type: Optional[PromptType] = None


class RequestRetrieveResponse(ApiBaseModel):
    results: list[RequestRetrieveResults]
    total_count: int


class SystemPromptRetrieveResponse(ApiBaseModel):
    results: list[SystemPrompt]
    total_count: int


class SystemPromptCreateResponse(ApiBaseModel):
    result: SystemPrompt


class SystemPromptIdRetrieveResponse(ApiBaseModel):
    result: SystemPrompt


class SystemPromptIdUpdateResponse(ApiBaseModel):
    result: SystemPrompt


class _TextChatCreateRequest(ApiBaseModel):
    conversation_id: Optional[str] = None
    messages: Optional[list[BaseMessage]] = Field(None, min_length=1)
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    parent_id: Optional[str] = None
    prompt_id: Optional[str] = None
    prompt_template_id: Optional[str] = None
    trim_method: Optional[TrimMethod] = None
    use_conversation_parameters: Optional[bool] = None


class TextChatCreateResponse(ApiBaseModel):
    conversation_id: str
    created_at: Optional[AwareDatetime] = None
    id: Optional[str] = None
    input_parameters: Optional[dict[str, Any]] = None
    model_id: Optional[str] = None
    results: list[TextGenerationResult]


class _TextChatOutputCreateRequest(ApiBaseModel):
    conversation_id: Optional[str] = None
    messages: Optional[list[BaseMessage]] = Field(None, min_length=1)
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    parent_id: Optional[str] = None
    prompt_id: Optional[str] = None
    prompt_template_id: Optional[str] = None
    trim_method: Optional[TrimMethod] = None
    use_conversation_parameters: Optional[bool] = None


class _TextChatStreamCreateRequest(ApiBaseModel):
    conversation_id: Optional[str] = None
    messages: Optional[list[BaseMessage]] = Field(None, min_length=1)
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    parent_id: Optional[str] = None
    prompt_id: Optional[str] = None
    prompt_template_id: Optional[str] = None
    trim_method: Optional[TrimMethod] = None
    use_conversation_parameters: Optional[bool] = None


class TextChatStreamCreateResponse(ApiBaseModel):
    conversation_id: str
    created_at: Optional[AwareDatetime] = None
    id: Optional[str] = None
    input_parameters: Optional[dict[str, Any]] = None
    model_id: Optional[str] = None
    moderation: Optional[TextCreateResponseModeration] = None
    results: Optional[list[TextGenerationStreamResult]] = None


class _TextGenerationCreateRequest(ApiBaseModel):
    data: Optional[PromptTemplateData] = None
    input: Optional[str] = Field(None, examples=["How are you"], title="Input string")
    """
    The input is the prompt to generate completions, passed as a string. Note: The method tokenizes the input internally. It is recommended not to leave any trailing spaces.
    """
    model_id: Optional[str] = Field(None, title="Model ID")
    """
    The ID of the model or tune to be used for this request.
    """
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = Field(None, min_length=1, title="Saved prompt Id")


class TextGenerationCreateResponse(ApiBaseModel):
    created_at: AwareDatetime
    id: str
    input_parameters: Optional[dict[str, Any]] = None
    model_id: str
    results: list[TextGenerationResult]


class _TextGenerationOutputCreateRequest(ApiBaseModel):
    data: Optional[PromptTemplateData] = None
    input: Optional[str] = None
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = None
    use_default: Optional[bool] = None


class _TextGenerationStreamCreateRequest(ApiBaseModel):
    data: Optional[PromptTemplateData] = None
    input: Optional[str] = Field(None, examples=["How are you"], title="Input string")
    """
    The input is the prompt to generate completions, passed as a string. Note: The method tokenizes the input internally. It is recommended not to leave any trailing spaces.
    """
    model_id: Optional[str] = Field(None, title="Model ID")
    """
    The ID of the model or tune to be used for this request.
    """
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = Field(None, min_length=1, title="Saved prompt Id")


class TextGenerationStreamCreateResponse(ApiBaseModel):
    created_at: Optional[AwareDatetime] = None
    id: Optional[str] = None
    input_parameters: Optional[dict[str, Any]] = None
    model_id: str
    moderation: Optional[TextCreateResponseModeration] = None
    results: Optional[list[TextGenerationStreamResult]] = None


class TextModerationCreateResponse(ApiBaseModel):
    results: list[TextCreateResponseModeration]


class _TextTokenizationCreateRequest(ApiBaseModel):
    data: Optional[PromptTemplateData] = None
    input: Optional[Union[str, list[str]]] = None
    model_id: Optional[str] = None
    parameters: Optional[TextTokenizationParameters] = None
    prompt_id: Optional[str] = None


class UserRetrieveResponse(ApiBaseModel):
    result: UserResponseResult


class UserPatchResponse(ApiBaseModel):
    result: UserResponseResult


class PromptsResponseResult(ApiBaseModel):
    author: Optional[PromptResultAuthor] = None
    created_at: AwareDatetime
    data: Optional[dict[str, Any]] = None
    description: Optional[str] = None
    id: str
    input: Optional[str] = None
    messages: Optional[list[BaseMessage]] = None
    metadata: Optional[dict[str, Any]] = None
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    name: str
    output: Optional[str] = None
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = None
    public: Optional[bool] = None
    task: Optional[PromptResultTask] = None
    type: PromptType


class RequestChatConversationIdRetrieveResultsRequest(ApiBaseModel):
    conversation_id: Optional[str] = None
    messages: Optional[list[BaseMessage]] = Field(None, min_length=1)
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    parent_id: Optional[str] = None
    prompt_id: Optional[str] = None
    prompt_template_id: Optional[str] = None
    trim_method: Optional[TrimMethod] = None
    use_conversation_parameters: Optional[bool] = None


class RequestChatConversationIdRetrieveResultsResponse(ApiBaseModel):
    conversation_id: str
    created_at: Optional[AwareDatetime] = None
    id: Optional[str] = None
    input_parameters: Optional[dict[str, Any]] = None
    model_id: Optional[str] = None
    results: list[TextGenerationResult]


class TextGenerationComparisonCreateRequestRequest(ApiBaseModel):
    data: Optional[PromptTemplateData] = None
    input: str
    model_id: Optional[str] = None
    moderations: Optional[ModerationParameters] = None
    parameters: Optional[TextGenerationParameters] = None
    prompt_id: Optional[str] = None
    use_default: Optional[bool] = None


class TextGenerationComparisonCreateResultsResult(ApiBaseModel):
    created_at: AwareDatetime
    id: str
    input_parameters: Optional[dict[str, Any]] = None
    model_id: str
    results: list[TextGenerationResult]


class TuneResult(ApiBaseModel):
    created_at: AwareDatetime
    datapoints: Optional[TuneResultDatapoints] = None
    evaluation_files: Optional[list[TuneResultFiles]] = None
    finished_at: Optional[AwareDatetime] = None
    id: str
    model_id: str
    model_name: str
    name: str
    parameters: Optional[dict[str, Any]] = None
    preferred: bool
    started_at: Optional[AwareDatetime] = None
    status: TuneStatus
    status_message: Optional[str] = None
    task_id: str
    task_name: str
    training_files: Optional[list[TuneResultFiles]] = None
    tuning_type: str
    validation_files: Optional[list[TuneResultFiles]] = None
    vectors: Optional[str] = None


class UserCreateResult(ApiBaseModel):
    api_key: UserCreateResultApiKey
    data_usage_consent: bool
    first_name: Optional[str] = None
    generate_default: Optional[UserGenerationDefault] = None
    id: int
    last_name: Optional[str] = None
    tou_accepted: bool
    tou_accepted_at: Optional[str] = None
    user_id: str


class PromptRetrieveResponse(ApiBaseModel):
    results: list[PromptsResponseResult]
    total_count: int


class PromptCreateResponse(ApiBaseModel):
    result: PromptsResponseResult


class PromptIdRetrieveResponse(ApiBaseModel):
    result: PromptsResponseResult


class PromptIdPatchResponse(ApiBaseModel):
    result: PromptsResponseResult


class PromptIdUpdateResponse(ApiBaseModel):
    result: PromptsResponseResult


class _TextGenerationComparisonCreateRequest(ApiBaseModel):
    compare_parameters: TextGenerationComparisonParameters
    name: Optional[str] = None
    request: TextGenerationComparisonCreateRequestRequest


class TuneRetrieveResponse(ApiBaseModel):
    results: list[TuneResult]
    total_count: int


class TuneCreateResponse(ApiBaseModel):
    result: TuneResult


class TuneImportCreateResponse(ApiBaseModel):
    result: TuneResult


class TuneIdRetrieveResponse(ApiBaseModel):
    result: TuneResult


class TuneIdPatchResponse(ApiBaseModel):
    result: TuneResult


class UserCreateResponse(ApiBaseModel):
    result: UserCreateResult


class RequestChatConversationIdRetrieveResults(ApiBaseModel):
    created_at: AwareDatetime
    duration: int
    id: str
    parent_id: Optional[str] = None
    request: RequestChatConversationIdRetrieveResultsRequest
    response: RequestChatConversationIdRetrieveResultsResponse
    status: RequestStatus
    version: Optional[RequestResultVersion] = None


class TextGenerationComparisonCreateResults(ApiBaseModel):
    error: Optional[Any] = None
    parameters: TextGenerationComparisonCreateResultsParameters
    result: Optional[TextGenerationComparisonCreateResultsResult] = None


class RequestChatConversationIdRetrieveResponse(ApiBaseModel):
    results: list[RequestChatConversationIdRetrieveResults]


class TextGenerationComparisonCreateResponse(ApiBaseModel):
    results: list[TextGenerationComparisonCreateResults]


__all__ = [
    "ApiKeyRegenerateCreateResponse",
    "ApiKeyResult",
    "ApiKeyRetrieveResponse",
    "BadRequestResponse",
    "BaseErrorExtension",
    "BaseErrorResponse",
    "BaseMessage",
    "BaseTokens",
    "ChatRole",
    "ConcurrencyLimit",
    "DecodingMethod",
    "Extensions",
    "Extensions1",
    "Extensions2",
    "Extensions3",
    "Extensions4",
    "Extensions5",
    "FileCreateResponse",
    "FileFormat",
    "FileIdRetrieveResponse",
    "FileListSortBy",
    "FilePurpose",
    "FileResult",
    "FileRetrieveResponse",
    "GeneratedToken",
    "HAPOptions",
    "ImplicitHateOptions",
    "Input",
    "InternalServerErrorResponse",
    "LengthPenalty",
    "ModelFamily",
    "ModelIdRetrieveResponse",
    "ModelIdRetrieveResult",
    "ModelRetrieveResponse",
    "ModelRetrieveResults",
    "ModelTokenLimits",
    "ModerationHAP",
    "ModerationImplicitHate",
    "ModerationParameters",
    "ModerationPosition",
    "ModerationStigma",
    "ModerationTokens",
    "NotFoundResponse",
    "PromptCreateResponse",
    "PromptIdPatchResponse",
    "PromptIdRetrieveResponse",
    "PromptIdUpdateResponse",
    "PromptResultAuthor",
    "PromptResultTask",
    "PromptRetrieveRequestParamsSource",
    "PromptRetrieveResponse",
    "PromptTemplate",
    "PromptTemplateData",
    "PromptType",
    "PromptsResponseResult",
    "RequestApiVersion",
    "RequestChatConversationIdRetrieveResponse",
    "RequestChatConversationIdRetrieveResults",
    "RequestChatConversationIdRetrieveResultsRequest",
    "RequestChatConversationIdRetrieveResultsResponse",
    "RequestEndpoint",
    "RequestOrigin",
    "RequestResultVersion",
    "RequestRetrieveResponse",
    "RequestRetrieveResults",
    "RequestStatus",
    "SortDirection",
    "StigmaOptions",
    "StopReason",
    "StorageProviderLocation",
    "SystemPrompt",
    "SystemPromptAuthor",
    "SystemPromptCreateResponse",
    "SystemPromptIdRetrieveResponse",
    "SystemPromptIdUpdateResponse",
    "SystemPromptRetrieveResponse",
    "SystemPromptType",
    "TaskRetrieveResponse",
    "Tasks",
    "TextChatCreateResponse",
    "TextChatOutputCreateResponse",
    "TextChatStreamCreateResponse",
    "TextCreateResponseModeration",
    "TextEmbeddingCreateResponse",
    "TextEmbeddingLimit",
    "TextEmbeddingLimitRetrieveResponse",
    "TextEmbeddingParameters",
    "TextGenerationComparisonCreateRequestRequest",
    "TextGenerationComparisonCreateResponse",
    "TextGenerationComparisonCreateResults",
    "TextGenerationComparisonCreateResultsParameters",
    "TextGenerationComparisonCreateResultsResult",
    "TextGenerationComparisonParameters",
    "TextGenerationCreateResponse",
    "TextGenerationFeedbackCategory",
    "TextGenerationFeedbackResult",
    "TextGenerationIdFeedbackCreateResponse",
    "TextGenerationIdFeedbackRetrieveResponse",
    "TextGenerationIdFeedbackUpdateResponse",
    "TextGenerationLimit",
    "TextGenerationLimitRetrieveResponse",
    "TextGenerationOutputCreateResponse",
    "TextGenerationParameters",
    "TextGenerationResult",
    "TextGenerationReturnOptions",
    "TextGenerationStreamCreateResponse",
    "TextGenerationStreamResult",
    "TextModeration",
    "TextModerationCreateResponse",
    "TextTokenizationCreateResponse",
    "TextTokenizationCreateResults",
    "TextTokenizationParameters",
    "TextTokenizationReturnOptions",
    "TooManyRequestsResponse",
    "TrimMethod",
    "TuneAssetType",
    "TuneCreateResponse",
    "TuneIdPatchResponse",
    "TuneIdRetrieveResponse",
    "TuneImportCreateResponse",
    "TuneListSortBy",
    "TuneParameters",
    "TuneResult",
    "TuneResultDatapointLoss",
    "TuneResultDatapoints",
    "TuneResultFiles",
    "TuneRetrieveResponse",
    "TuneStatus",
    "TunesResultDatapointLoss",
    "TuningType",
    "TuningTypeRetrieveResponse",
    "TuningTypeRetrieveResults",
    "UnauthorizedResponse",
    "UnavailableResponse",
    "UserCreateResponse",
    "UserCreateResult",
    "UserCreateResultApiKey",
    "UserGenerationDefault",
    "UserPatchResponse",
    "UserResponseResult",
    "UserRetrieveResponse",
]



"""Extension for LLamaIndex library"""

from genai.extensions.llama_index.embeddings import IBMGenAILlamaIndexEmbedding
from genai.extensions.llama_index.llm import IBMGenAILlamaIndex

__all__ = ["IBMGenAILlamaIndex", "IBMGenAILlamaIndexEmbedding"]



from __future__ import annotations

from typing import Optional, overload

from genai._types import ModelLike
from genai._utils.api_client import ApiClient
from genai._utils.api_client import BaseConfig as ApiClientConfig
from genai._utils.general import to_model_instance
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
)
from genai.credentials import Credentials
from genai.file import FileService as _FileService
from genai.model import ModelService as _ModelService
from genai.prompt import PromptService as _PromptService
from genai.request import RequestService as _RequestService
from genai.system_prompt import SystemPromptService as _SystemPromptService
from genai.text import TextService as _TextService
from genai.tune import TuneService as _TuneService
from genai.user import UserService as _UserService

__all__ = ["Client", "BaseConfig", "BaseServices"]


class BaseServices(BaseServiceServices):
    """Appropriate services used by the Client"""

    TextService: type[_TextService] = _TextService
    RequestService: type[_RequestService] = _RequestService
    TuneService: type[_TuneService] = _TuneService
    ModelService: type[_ModelService] = _ModelService
    FileService: type[_FileService] = _FileService
    PromptService: type[_PromptService] = _PromptService
    SystemPromptService: type[_SystemPromptService] = _SystemPromptService
    UserService: type[_UserService] = _UserService


class BaseConfig(BaseServiceConfig):
    """Client's configuration model"""

    api_client_config: Optional[ModelLike[ApiClientConfig]] = None


class Client(BaseService[BaseConfig, BaseServices]):
    """
    The `Client` class provides an interface for interacting with various services through an API client.
    It can be initialized with either an `api_client` or `credentials` along with optional configurations and services.

    Example::

        from genai import Credentials, Client

        credentials = Credentials.from_env()
        client = Client(credentials=credentials)

    Attributes:
        text: An instance of the `TextService` class for text-related operations.
        request: An instance of the `RequestService` class for making request-related operations.
        tune: An instance of the `TuneService` class for tuning models.
        model: An instance of the `ModelService` class for managing models.
        file: An instance of the `FileService` class for managing files.
        prompt: An instance of the `PromptService` class for working with prompts.
        system_prompt: An instance of the `SystemPromptService` class for working with system prompts.
        user: An instance of the `UserService` class for managing user-related operations.
    """

    Config = BaseConfig
    Services = BaseServices

    @overload
    def __init__(
        self,
        *,
        api_client: ApiClient,
        config: Optional[ModelLike[BaseConfig]] = None,
        services: Optional[BaseServices] = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        *,
        credentials: Credentials,
        config: Optional[ModelLike[BaseConfig]] = None,
        services: Optional[BaseServices] = None,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        credentials: Optional[Credentials] = None,
        api_client: Optional[ApiClient] = None,
        config: Optional[ModelLike[BaseConfig]] = None,
        services: Optional[BaseServices] = None,
    ) -> None:
        """
        Args:
            credentials: The credentials used to authenticate the API client.
            api_client: The API client used to make requests to the API.
            config: The configuration for the API client.
            services: The services object containing instances of various service classes.

        Raises:
            ValueError: Either 'api_client' or 'credentials' needs to be passed.

        Note:
            The `api_client` parameter must be provided either directly or through `credentials`.
            If no `services` parameter is provided, the client will use the default one.
        """
        config = to_model_instance(config, self.Config)

        if (not api_client and not credentials) or (api_client and credentials):
            raise ValueError("Either 'api_client' or 'credentials' needs to passed.")

        if not api_client:
            assert credentials
            api_client = ApiClient(credentials=credentials, config=config.api_client_config)

        assert api_client and config
        super().__init__(api_client=api_client, config=config)

        if not services:
            services = self.Services()

        self.text = services.TextService(api_client=api_client)
        self.request = services.RequestService(api_client=api_client)
        self.tune = services.TuneService(api_client=api_client)
        self.model = services.ModelService(api_client=api_client)
        self.file = services.FileService(api_client=api_client)
        self.prompt = services.PromptService(api_client=api_client)
        self.system_prompt = services.SystemPromptService(api_client=api_client)
        self.user = services.UserService(api_client=api_client)



"""Wrapper around IBM GENAI APIs for use in Langchain"""

import logging
from pathlib import Path
from typing import Any, Iterator, Optional, Union

from pydantic import ConfigDict
from pydantic.v1 import validator

from genai import Client
from genai._types import EnumLike
from genai._utils.general import to_model_optional
from genai.extensions._common.utils import (
    _prepare_chat_generation_request,
    create_generation_info_from_response,
)
from genai.schema import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    ModerationParameters,
    SystemMessage,
    TextGenerationParameters,
    TrimMethod,
)

try:
    from langchain_core.callbacks.manager import CallbackManagerForLLMRun
    from langchain_core.language_models.chat_models import BaseChatModel
    from langchain_core.messages import AIMessage as LCAIMessage
    from langchain_core.messages import AIMessageChunk as LCAIMessageChunk
    from langchain_core.messages import BaseMessage as LCBaseMessage
    from langchain_core.messages import ChatMessage as LCChatMessage
    from langchain_core.messages import HumanMessage as LCHumanMessage
    from langchain_core.messages import SystemMessage as LCSystemMessage
    from langchain_core.messages import get_buffer_string
    from langchain_core.outputs import ChatGeneration, ChatResult

    from genai.extensions.langchain.utils import (
        CustomChatGenerationChunk,
        create_llm_output,
        dump_optional_model,
        load_config,
        update_token_usage_stream,
    )
except ImportError:
    raise ImportError("Could not import langchain: Please install ibm-generative-ai[langchain] extension.")  # noqa: B904

__all__ = ["LangChainChatInterface"]

logger = logging.getLogger(__name__)

Message = Union[LCBaseMessage, BaseMessage]
Messages = Union[list[LCBaseMessage], list[Message]]


def _convert_message_to_genai(message: Message) -> BaseMessage:
    def convert_message_content(content: Any) -> str:
        if not isinstance(content, str):
            raise TypeError(
                f"Cannot convert non-string message content. Got {content} of type {type(content)}, expected string."
            )

        return content

    if isinstance(message, BaseMessage):
        return message
    elif isinstance(message, LCChatMessage) or isinstance(message, LCHumanMessage):
        return HumanMessage(content=convert_message_content(message.content))
    elif isinstance(message, LCAIMessage):
        return AIMessage(content=convert_message_content(message.content))
    elif isinstance(message, LCSystemMessage):
        return SystemMessage(content=convert_message_content(message.content))
    else:
        raise ValueError(f"Got unknown message type '{message}'")


def _convert_messages_to_genai(messages: Messages) -> list[BaseMessage]:
    return [_convert_message_to_genai(msg) for msg in messages]


class LangChainChatInterface(BaseChatModel):
    """
    Class representing the LangChainChatInterface for interacting with the LangChain chat API.

    Example::

        from genai import Client, Credentials
        from genai.extensions.langchain import LangChainChatInterface
        from langchain_core.messages import HumanMessage, SystemMessage
        from genai.schema import TextGenerationParameters

        client = Client(credentials=Credentials.from_env())
        llm = LangChainChatInterface(
            client=client,
            model_id="meta-llama/llama-2-70b-chat",
            parameters=TextGenerationParameters(
                max_new_tokens=250,
            )
        )

        response = chat_model.generate(messages=[HumanMessage(content="Hello world!")])
        print(response)

    """

    model_config = ConfigDict(extra="forbid", protected_namespaces=())

    client: Client
    model_id: str
    prompt_id: Optional[str] = None
    parameters: Optional[TextGenerationParameters] = None
    moderations: Optional[ModerationParameters] = None
    parent_id: Optional[str] = None
    prompt_template_id: Optional[str] = None
    trim_method: Optional[EnumLike[TrimMethod]] = None
    use_conversation_parameters: Optional[bool] = None
    conversation_id: Optional[str] = None
    streaming: Optional[bool] = None

    @validator("parameters", "moderations", pre=True, always=True)
    @classmethod
    def validate_data_models(cls, value, values, config, field):
        return to_model_optional(value, Model=field.type_, copy=False)

    @classmethod
    def is_lc_serializable(cls) -> bool:
        return True

    @property
    def lc_secrets(self) -> dict[str, str]:
        return {"client": "CLIENT"}

    @classmethod
    def load_from_file(cls, file: Union[str, Path], *, client: Client):
        config = load_config(file)
        return cls(**config, client=client)

    @property
    def _identifying_params(self) -> dict[str, Any]:
        return {
            "model_id": self.model_id,
            "prompt_id": self.prompt_id,
            "parameters": dump_optional_model(self.parameters),
            "moderations": dump_optional_model(self.moderations),
            "parent_id": self.parent_id,
            "prompt_template_id": self.prompt_template_id,
            "trim_method": self.trim_method,
            "use_conversation_parameters": self.use_conversation_parameters,
            **super()._identifying_params,
        }

    @property
    def _llm_type(self) -> str:
        return "ibmgenai_chat_llm"

    def _prepare_request(self, **kwargs):
        updated = {k: kwargs.pop(k, v) for k, v in self._identifying_params.items()}
        return _prepare_chat_generation_request(**kwargs, **updated)

    def _stream(
        self,
        messages: Messages,
        stop: Optional[list[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[CustomChatGenerationChunk]:
        for response in self.client.text.chat.create_stream(
            **self._prepare_request(messages=_convert_messages_to_genai(messages), stop=stop, **kwargs)
        ):
            if not response:
                continue

            def send_chunk(*, text: str = "", generation_info: dict):
                logger.info("Chunk received: {}".format(text))
                chunk = CustomChatGenerationChunk(
                    message=LCAIMessageChunk(content=text, generation_info=generation_info),
                    generation_info=generation_info,
                )
                yield chunk
                if run_manager:
                    run_manager.on_llm_new_token(token=text, chunk=chunk, response=response)  # noqa: B023
                    # Function definition does not bind loop variable `response`: linter is probably just confused here

            if response.moderation:
                generation_info = create_generation_info_from_response(response, result=response.moderation)
                yield from send_chunk(generation_info=generation_info)

            for result in response.results or []:
                generation_info = create_generation_info_from_response(response, result=result)
                yield from send_chunk(text=result.generated_text or "", generation_info=generation_info)

    def _generate(
        self,
        messages: Messages,
        stop: Optional[list[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        def handle_stream():
            final_generation: Optional[CustomChatGenerationChunk] = None
            for result in self._stream(
                messages=messages,
                stop=stop,
                run_manager=run_manager,
                **kwargs,
            ):
                if final_generation:
                    token_usage = result.generation_info.pop("token_usage")
                    final_generation += result
                    update_token_usage_stream(
                        target=final_generation.generation_info["token_usage"],
                        source=token_usage,
                    )
                else:
                    final_generation = result

            assert final_generation and final_generation.generation_info
            return {
                "text": final_generation.text,
                "generation_info": final_generation.generation_info.copy(),
            }

        def handle_non_stream():
            response = self.client.text.chat.create(
                **self._prepare_request(messages=_convert_messages_to_genai(messages), stop=stop, **kwargs),
            )

            assert response.results
            result = response.results[0]

            return {
                "text": result.generated_text or "",
                "generation_info": create_generation_info_from_response(response, result=result),
            }

        result = handle_stream() if self.streaming else handle_non_stream()
        return ChatResult(
            generations=[
                ChatGeneration(
                    message=LCAIMessage(content=result["text"]),
                    generation_info=result["generation_info"].copy(),
                )
            ],
            llm_output=create_llm_output(
                model=result["generation_info"].get("model_id", self.model_id or ""),
                token_usages=[result["generation_info"]["token_usage"]],
            ),
        )

    def get_num_tokens(self, text: str) -> int:
        response = list(self.client.text.tokenization.create(model_id=self.model_id, input=[text]))[0]
        return response.results[0].token_count

    def get_num_tokens_from_messages(self, messages: list[LCBaseMessage]) -> int:
        return sum(
            sum(result.token_count for result in response.results)
            for response in self.client.text.tokenization.create(
                model_id=self.model_id, input=[get_buffer_string([message]) for message in messages]
            )
        )

    def _combine_llm_outputs(self, llm_outputs: list[Optional[dict]]) -> dict:
        token_usages: list[Optional[dict]] = []
        model = ""

        for output in llm_outputs:
            if output:
                model = model or output.get("meta", {}).get("model_id")
                token_usages.append(output.get("token_usage"))

        return create_llm_output(model=model or self.model_id, token_usages=token_usages)

    def get_token_ids(self, text: str) -> list[int]:
        raise NotImplementedError("API does not support returning token ids.")



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, ConfigDict

from genai._types import ModelLike
from genai.client import Client
from genai.schema import TextEmbeddingParameters
from genai.text.embedding.embedding_service import CreateExecutionOptions

__all__ = ["LangChainEmbeddingsInterface"]

try:
    from langchain_core.embeddings import Embeddings
    from langchain_core.runnables.config import run_in_executor
except ImportError:
    raise ImportError("Could not import langchain: Please install ibm-generative-ai[langchain] extension.")  # noqa: B904


class LangChainEmbeddingsInterface(BaseModel, Embeddings):
    """
    Class representing the LangChainChatInterface for interacting with the LangChain chat API.

    Example::

        from genai import Client, Credentials
        from genai.extensions.langchain import LangChainEmbeddingsInterface
        from genai.text.embedding import TextEmbeddingParameters

        client = Client(credentials=Credentials.from_env())
        embeddings = LangChainEmbeddingsInterface(
            client=client,
            model_id="sentence-transformers/all-minilm-l6-v2",
            parameters=TextEmbeddingParameters(truncate_input_tokens=True)
        )

        embeddings.embed_query("Hello world!")
        embeddings.embed_documents(["First document", "Second document"])
    """

    model_config = ConfigDict(extra="forbid", protected_namespaces=(), arbitrary_types_allowed=True)

    client: Client
    model_id: str
    parameters: Optional[ModelLike[TextEmbeddingParameters]] = None
    execution_options: Optional[ModelLike[CreateExecutionOptions]] = None

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Embed search documents"""
        return self._get_embeddings(texts)

    def embed_query(self, text: str) -> list[float]:
        """Embed query text."""
        response = self._get_embeddings([text])
        return response[0]

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous Embed query text."""
        return await run_in_executor(None, self.embed_query, text)

    async def aembed_documents(self, texts: List[str]) -> list[list[float]]:
        """Asynchronous Embed search documents"""
        return await run_in_executor(None, self.embed_documents, texts)

    def _get_embeddings(self, texts: list[str]) -> list[list[float]]:
        embeddings: list[list[float]] = []
        for response in self.client.text.embedding.create(
            model_id=self.model_id, inputs=texts, parameters=self.parameters, execution_options=self.execution_options
        ):
            embeddings.extend(response.results)

        return embeddings


LangChainEmbeddingsInterface.model_rebuild()
